import uuid
from dataclasses import dataclass, field
from functools import partial

from typing import Optional, Dict, Set, List, Tuple

from skyplane import compute
from skyplane.utils import logger
from skyplane.utils.fn import do_parallel
from skyplane.utils.timer import Timer


@dataclass
class ProvisionerTask:
    """Dataclass to track a single VM provisioning job."""

    cloud_provider: str
    region: str
    vm_type: str
    spot: bool = False
    autoterminate_minutes: Optional[int] = None
    tags: Dict[str, str] = field(default_factory=dict)
    uuid: str = field(default_factory=lambda: str(uuid.uuid4()))

    def __hash__(self):
        return uuid.UUID(self.uuid).int


class Provisioner:
    def __init__(
        self,
        aws_auth: Optional[compute.AWSAuthentication] = None,
        azure_auth: Optional[compute.AzureAuthentication] = None,
        gcp_auth: Optional[compute.GCPAuthentication] = None,
        host_uuid: Optional[str] = None,
    ):
        self.aws_auth = aws_auth
        self.azure_auth = azure_auth
        self.gcp_auth = gcp_auth
        self.host_uuid = host_uuid
        self._make_cloud_providers()
        self.temp_nodes: Set[compute.Server] = set()  # temporary area to store nodes that should be terminated upon exit
        self.pending_provisioner_tasks: List[ProvisionerTask] = []
        self.provisioned_vms: Dict[str, compute.Server] = {}

        # store GCP firewall rules to be deleted upon exit
        self.gcp_firewall_rules: Set[str] = set()

    def _make_cloud_providers(self):
        self.aws = compute.AWSCloudProvider(
            key_prefix=f"skyplane{'-'+self.host_uuid.replace('-', '') if self.host_uuid else ''}", auth=self.aws_auth
        )
        self.azure = compute.AzureCloudProvider(auth=self.azure_auth)
        self.gcp = compute.GCPCloudProvider(auth=self.gcp_auth)

    def init_global(self, aws: bool = True, azure: bool = True, gcp: bool = True):
        logger.fs.info(f"[Provisioner.init_global] Initializing global resources for aws={aws}, azure={azure}, gcp={gcp}")
        jobs = []
        if aws:
            jobs.append(partial(self.aws.setup_global, attach_policy_arn="arn:aws:iam::aws:policy/AmazonS3FullAccess"))
        if azure:
            jobs.append(self.azure.create_ssh_key)
            jobs.append(self.azure.set_up_resource_group)
        if gcp:
            jobs.append(self.gcp.setup_global)
        do_parallel(lambda fn: fn(), jobs, spinner=False)

    def add_task(
        self,
        cloud_provider: str,
        region: str,
        vm_type: str,
        spot: bool = False,
        autoterminate_minutes: Optional[int] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        if tags is None:
            tags = {"skyplane": "true", "skyplaneclientid": self.host_uuid} if self.host_uuid else {"skyplane": "true"}
        task = ProvisionerTask(cloud_provider, region, vm_type, spot, autoterminate_minutes, tags)
        self.pending_provisioner_tasks.append(task)
        logger.fs.info(f"[Provisioner.add_task] Queue {task}")
        return task.uuid

    def get_node(self, uuid: str) -> compute.Server:
        return self.provisioned_vms[uuid]

    def _provision_task(self, task: ProvisionerTask):
        with Timer() as t:
            if task.cloud_provider == "aws":
                assert self.aws.auth.enabled(), "AWS credentials not configured"
                server = self.aws.provision_instance(task.region, task.vm_type, use_spot_instances=task.spot, tags=task.tags)
            elif task.cloud_provider == "azure":
                assert self.azure.auth.enabled(), "Azure credentials not configured"
                server = self.azure.provision_instance(task.region, task.vm_type, use_spot_instances=task.spot, tags=task.tags)
            elif task.cloud_provider == "gcp":
                assert self.gcp.auth.enabled(), "GCP credentials not configured"
                # todo specify network tier in ReplicationTopology
                server = self.gcp.provision_instance(
                    task.region,
                    task.vm_type,
                    use_spot_instances=task.spot,
                    gcp_premium_network=False,
                    tags=task.tags,
                )
            else:
                raise NotImplementedError(f"Unknown provider {task.cloud_provider}")
        logger.fs.debug(f"[Provisioner._provision_task] Provisioned {server} in {t.elapsed:.2f}s")
        self.temp_nodes.add(server)
        self.provisioned_vms[task.uuid] = server
        server.wait_for_ssh_ready()
        if task.autoterminate_minutes:
            server.enable_auto_shutdown(task.autoterminate_minutes)
        self.temp_nodes.remove(server)
        return server

    def provision(self, authorize_firewall: bool = True, max_jobs: int = 16, spinner: bool = False) -> List[str]:
        """Provision the VMs in the pending_provisioner_tasks list. Returns UUIDs of provisioned VMs."""
        if not self.pending_provisioner_tasks:
            return []

        # copy list to avoid concurrency issue
        provision_tasks = list(self.pending_provisioner_tasks)
        aws_provisioned = any([task.cloud_provider == "aws" for task in provision_tasks])
        aws_regions = set([task.region for task in provision_tasks if task.cloud_provider == "aws"])
        azure_provisioned = any([task.cloud_provider == "azure" for task in provision_tasks])
        gcp_provisioned = any([task.cloud_provider == "gcp" for task in provision_tasks])

        # configure regions
        if aws_provisioned:
            do_parallel(
                self.aws.setup_region, list(set(aws_regions)), spinner=spinner, spinner_persist=False, desc="Configuring AWS regions"
            )
            logger.fs.info(f"[Provisioner.provision] Configured AWS regions {aws_regions}")

        # provision VMs
        logger.fs.info(f"[Provisioner.provision] Provisioning {len(provision_tasks)} VMs")
        results: List[Tuple[ProvisionerTask, compute.Server]] = do_parallel(
            self._provision_task,
            provision_tasks,
            n=max_jobs,
            spinner=spinner,
            spinner_persist=spinner,
            desc="Provisioning VMs",
        )

        # configure firewall
        if authorize_firewall:
            public_ips = [s.public_ip() for _, s in results]
            # authorize access to private IPs for GCP VMs due to global VPC
            private_ips = [s.private_ip() for t, s in results if t.cloud_provider == "gcp"]
            authorize_ip_jobs = []
            if aws_provisioned:
                authorize_ip_jobs.extend([partial(self.aws.add_ips_to_security_group, r, public_ips) for r in set(aws_regions)])
            if gcp_provisioned:

                def authorize_gcp_gateways():
                    self.gcp_firewall_rules.add(self.gcp.authorize_gateways(public_ips + private_ips))

                authorize_ip_jobs.append(authorize_gcp_gateways)

            do_parallel(
                lambda fn: fn(),
                authorize_ip_jobs,
                n=max_jobs,
                spinner=spinner,
                spinner_persist=False,
                desc="Authorizing gateways with firewalls",
            )
            logger.fs.info(f"[Provisioner.provision] Authorized AWS gateways with firewalls: {public_ips}")
            logger.fs.info(f"[Provisioner.provision] Authorized GCP gateways with firewalls: {public_ips}, {private_ips}")

        for task in provision_tasks:
            self.pending_provisioner_tasks.remove(task)

        return [task.uuid for task in provision_tasks]

    def deprovision(self, deauthorize_firewall: bool = True, max_jobs: int = 64, spinner: bool = False):
        """Deprovision all nodes. Returns UUIDs of deprovisioned VMs."""
        if not self.provisioned_vms and not self.temp_nodes:
            return []

        def deprovision_gateway_instance(server: compute.Server):
            server.terminate_instance()
            idx_to_del = None
            for idx, s in list(self.provisioned_vms.items()):
                if s == server:
                    idx_to_del = idx
                    break
            if idx_to_del:
                del self.provisioned_vms[idx_to_del]
            else:
                logger.fs.warning(f"[Provisioner.deprovision] Could not find {server} in {self.provisioned_vms}")
            if server in self.temp_nodes:
                self.temp_nodes.remove(server)
            logger.fs.info(f"[Provisioner.deprovision] Terminated {server}")

        servers = list(self.provisioned_vms.values()) + list(self.temp_nodes)
        aws_deprovisioned = any([s.provider == "aws" for s in servers])
        azure_deprovisioned = any([s.provider == "azure" for s in servers])
        gcp_deprovisioned = any([s.provider == "gcp" for s in servers])
        if azure_deprovisioned:
            logger.warning("Azure deprovisioning is very slow. Please be patient.")
        logger.fs.info(f"[Provisioner.deprovision] Deprovisioning {len(servers)} VMs")
        do_parallel(
            deprovision_gateway_instance,
            servers,
            n=max_jobs,
            spinner=spinner,
            spinner_persist=False,
            desc="Deprovisioning VMs",
        )

        # clean up firewall
        if deauthorize_firewall:
            # todo remove firewall rules for Azure
            public_ips = [s.public_ip() for s in servers]
            jobs = []
            if aws_deprovisioned:
                aws_regions = set([s.region() for s in servers if s.provider == "aws"])
                jobs.extend([partial(self.aws.remove_ips_from_security_group, r, public_ips) for r in set(aws_regions)])
                logger.fs.info(f"[Provisioner.deprovision] Deauthorizing AWS gateways with firewalls: {public_ips}")
            if gcp_deprovisioned:
                jobs.extend([partial(self.gcp.remove_gateway_rule, rule) for rule in self.gcp_firewall_rules])
                logger.fs.info(f"[Provisioner.deprovision] Deauthorizing GCP gateways with firewalls: {self.gcp_firewall_rules}")
            do_parallel(
                lambda fn: fn(), jobs, n=max_jobs, spinner=spinner, spinner_persist=False, desc="Deauthorizing gateways from firewalls"
            )



import json
import shutil
from dataclasses import dataclass

from typing import Dict, List, Optional, Set, Tuple

from skyplane.utils import logger


@dataclass
class ReplicationTopologyNode:
    region: str

    def to_dict(self) -> Dict:
        """Serialize to dict with type information."""
        return {"type": self.__class__.__name__, "fields": self.__dict__}

    @classmethod
    def from_dict(cls, topology_dict: Dict) -> "ReplicationTopologyNode":
        """Deserialize from dict with type information."""
        if topology_dict["type"] == "ReplicationTopologyGateway":
            return ReplicationTopologyGateway.from_dict_fields(topology_dict["fields"])
        elif topology_dict["type"] == "ReplicationTopologyObjectStore":
            return ReplicationTopologyObjectStore.from_dict_fields(topology_dict["fields"])
        else:
            raise ValueError("Unknown topology node type: {}".format(topology_dict["type"]))

    @classmethod
    def from_dict_fields(cls, fields: Dict):
        """Deserialize from dict with type information."""
        return cls(**fields)


@dataclass
class ReplicationTopologyGateway(ReplicationTopologyNode):
    instance: int

    def __hash__(self) -> int:
        return hash((self.region, self.instance))


@dataclass
class ReplicationTopologyObjectStore(ReplicationTopologyNode):
    def __hash__(self) -> int:
        return hash(self.region)


class ReplicationTopology:
    """
    ReplicationTopology stores a DAG where nodes are an instance in a cloud region
    (e.g. "aws:us-east-1", instance 0) and edges denote a connection to another
    cloud region (e.g. ("aws:us-east-1", 0) -> ("aws:us-west-2", 1) with an
    associated number of connections (e.g. 64).
    """

    def __init__(
        self,
        edges: Optional[List[Tuple[ReplicationTopologyNode, ReplicationTopologyNode, int]]] = None,
        cost_per_gb: Optional[float] = None,
    ):
        self.edges: List[Tuple[ReplicationTopologyNode, ReplicationTopologyNode, int]] = edges or []
        self.nodes: Set[ReplicationTopologyNode] = set(k[0] for k in self.edges) | set(k[1] for k in self.edges)
        self.cost_per_gb: Optional[float] = cost_per_gb

    @property
    def gateway_nodes(self) -> Set[ReplicationTopologyGateway]:
        return {n for n in self.nodes if isinstance(n, ReplicationTopologyGateway)}

    @property
    def obj_store_nodes(self) -> Set[ReplicationTopologyObjectStore]:
        return {n for n in self.nodes if isinstance(n, ReplicationTopologyObjectStore)}

    def add_instance_instance_edge(self, src_region: str, src_instance: int, dest_region: str, dest_instance: int, num_connections: int):
        """Add relay edge between two instances."""
        src_gateway = ReplicationTopologyGateway(src_region, src_instance)
        dest_gateway = ReplicationTopologyGateway(dest_region, dest_instance)
        self.edges.append((src_gateway, dest_gateway, int(num_connections)))
        self.nodes.add(src_gateway)
        self.nodes.add(dest_gateway)

    def add_objstore_instance_edge(self, src_region: str, dest_region: str, dest_instance: int):
        """Add object store to instance node (i.e. source bucket to source gateway)."""
        src_objstore = ReplicationTopologyObjectStore(src_region)
        dest_gateway = ReplicationTopologyGateway(dest_region, dest_instance)
        self.edges.append((src_objstore, dest_gateway, 0))
        self.nodes.add(src_objstore)
        self.nodes.add(dest_gateway)

    def add_instance_objstore_edge(self, src_region: str, src_instance: int, dest_region: str):
        """Add instance to object store edge (i.e. destination gateway to destination bucket)."""
        src_gateway = ReplicationTopologyGateway(src_region, src_instance)
        dest_objstore = ReplicationTopologyObjectStore(dest_region)
        self.edges.append((src_gateway, dest_objstore, 0))
        self.nodes.add(src_gateway)
        self.nodes.add(dest_objstore)

    def get_outgoing_paths(self, src: ReplicationTopologyNode):
        """Return nodes that follow src in the topology."""
        return {dest_gateway: num_connections for src_gateway, dest_gateway, num_connections in self.edges if src_gateway == src}

    def get_incoming_paths(self, dest: ReplicationTopologyNode):
        """Return nodes that precede dest in the topology."""
        return {src_gateway: num_connections for dest_gateway, src_gateway, num_connections in self.edges if dest_gateway == dest}

    def source_instances(self) -> Set[ReplicationTopologyGateway]:
        nodes = self.nodes - {v for u, v, _ in self.edges if not isinstance(u, ReplicationTopologyObjectStore)}
        return {n for n in nodes if isinstance(n, ReplicationTopologyGateway)}

    def sink_instances(self) -> Set[ReplicationTopologyGateway]:
        nodes = self.nodes - {u for u, v, _ in self.edges if not isinstance(v, ReplicationTopologyObjectStore)}
        return {n for n in nodes if isinstance(n, ReplicationTopologyGateway)}

    def source_region(self) -> str:
        instances = list(self.source_instances())
        assert all(
            i.region == instances[0].region for i in instances
        ), f"All source instances must be in the same region, but found {instances}"
        return instances[0].region

    def sink_region(self) -> str:
        instances = list(self.sink_instances())
        assert all(i.region == instances[0].region for i in instances), "All sink instances must be in the same region"
        return instances[0].region

    def per_region_count(self) -> Dict[str, int]:
        counts = {}
        for node in self.nodes:
            if isinstance(node, ReplicationTopologyGateway):
                counts[node.region] = counts.get(node.region, 0) + 1
        return counts

    def to_json(self):
        """
        Returns a JSON representation of the topology.
        """
        edges = []
        for e in self.edges:
            edges.append({"src": e[0].to_dict(), "dest": e[1].to_dict(), "num_connections": int(e[2])})
        return json.dumps(dict(replication_topology_edges=edges))

    @classmethod
    def from_json(cls, json_str: str):
        """
        Returns a ReplicationTopology from a JSON string.
        """
        in_dict = json.loads(json_str)
        assert "replication_topology_edges" in in_dict
        edges = []
        for edge in in_dict["replication_topology_edges"]:
            edges.append(
                (ReplicationTopologyNode.from_dict(edge["src"]), ReplicationTopologyNode.from_dict(edge["dest"]), edge["num_connections"])
            )
        return ReplicationTopology(edges)

    def to_graphviz(self):
        import graphviz as gv  # pytype: disable=import-error

        # if dot is not installed
        has_dot = shutil.which("dot") is not None
        if not has_dot:
            logger.error("Graphviz is not installed. Please install it to plot the solution (sudo apt install graphviz).")
            return None

        g = gv.Digraph(name="throughput_graph")
        g.attr(rankdir="LR")
        subgraphs = {}
        for src_gateway, dest_gateway, n_connections in self.edges:
            # group node instances by region
            src_region, src_instance = (
                src_gateway.region,
                src_gateway.instance if isinstance(src_gateway, ReplicationTopologyGateway) else "objstore",
            )
            dest_region, dest_instance = (
                dest_gateway.region,
                dest_gateway.instance if isinstance(dest_gateway, ReplicationTopologyGateway) else "objstore",
            )
            src_region, dest_region = src_region.replace(":", "/"), dest_region.replace(":", "/")
            src_node = f"{src_region}, {src_instance}"
            dest_node = f"{dest_region}, {dest_instance}"

            # make a subgraph for each region
            if src_region not in subgraphs:
                subgraphs[src_region] = gv.Digraph(name=f"cluster_{src_region}")
                subgraphs[src_region].attr(label=src_region)
            if dest_region not in subgraphs:
                subgraphs[dest_region] = gv.Digraph(name=f"cluster_{dest_region}")
                subgraphs[dest_region].attr(label=dest_region)

            # add nodes
            subgraphs[src_region].node(src_node, label=str(src_instance), shape="box")
            subgraphs[dest_region].node(dest_node, label=str(dest_instance), shape="box")

            # add edges
            g.edge(
                src_node,
                dest_node,
                label=f"{n_connections} connections" if src_instance != "objstore" and dest_instance != "objstore" else None,
            )

        for subgraph in subgraphs.values():
            g.subgraph(subgraph)

        return g



from functools import partial
from typing import Optional

from rich.console import Console

from skyplane import compute
from skyplane.utils import logger
from skyplane.utils.fn import do_parallel

console = Console()


def print_header():
    header = """ _____ _   ____   _______ _       ___   _   _  _____ 
/  ___| | / /\ \ / / ___ \ |     / _ \ | \ | ||  ___|
\ `--.| |/ /  \ V /| |_/ / |    / /_\ \|  \| || |__  
 `--. \    \   \ / |  __/| |    |  _  || . ` ||  __| 
/\__/ / |\  \  | | | |   | |____| | | || |\  || |___ 
\____/\_| \_/  \_/ \_|   \_____/\_| |_/\_| \_/\____/"""
    console.print(f"[bright_black]{header}[/bright_black]\n")


def print_stats_completed(total_runtime_s: float, throughput_gbits: Optional[float]):
    console.print(f"\n:white_check_mark: [bold green]Transfer completed successfully[/bold green]")
    runtime_line = f"[white]Transfer runtime:[/white] [bright_black]{total_runtime_s:.2f}s[/bright_black]"
    throughput_line = (
        f", [white]Throughput:[/white] [bright_black]{throughput_gbits:.2f}Gbps[/bright_black]" if throughput_gbits is not None else ""
    )
    console.print(f"{runtime_line}{throughput_line}")


def query_instances():
    instances = []
    query_jobs = []

    def catch_error(fn):
        def run():
            try:
                return fn()
            except Exception as e:
                logger.error(f"Error encountered during deprovision: {e}")
                return []

        return run

    if compute.AWSAuthentication().enabled():
        aws = compute.AWSCloudProvider()
        for region in aws.region_list():
            query_jobs.append(catch_error(partial(aws.get_matching_instances, region)))
    if compute.AzureAuthentication().enabled():
        query_jobs.append(catch_error(lambda: compute.AzureCloudProvider().get_matching_instances()))
    if compute.GCPAuthentication().enabled():
        query_jobs.append(catch_error(lambda: compute.GCPCloudProvider().get_matching_instances()))
    # query in parallel
    for instance_list in do_parallel(
        lambda f: f(), query_jobs, n=-1, return_args=False, spinner=True, desc="Querying clouds for instances"
    ):
        instances.extend(instance_list)
    return instances
